/***************************************************************************/
/***************************************************************************/
/*                                                                         */
/*   (c) 1995.  The Regents of the University of California.  All rights   */
/*   reserved.                                                             */
/*                                                                         */
/*   This work was produced at the University of California, Lawrence      */
/*   Livermore National Laboratory (UC LLNL) under contract no.            */
/*   W-7405-ENG-48 (Contract 48) between the U.S. Department of Energy     */
/*   (DOE) and The Regents of the University of California (University)    */
/*   for the operation of UC LLNL.  Copyright is reserved to the           */
/*   University for purposes of controlled dissemination,                  */
/*   commercialization through formal licensing, or other disposition      */
/*   under terms of Contract 48; DOE policies, regulations and orders;     */
/*   and U.S. statutes.  The rights of the Federal Government are          */
/*   reserved under Contract 48 subject to the restrictions agreed upon    */
/*   by the DOE and University.                                            */
/*                                                                         */
/*                                                                         */
/*                              DISCLAIMER                                 */
/*                                                                         */
/*   This software was prepared as an account of work sponsored by an      */
/*   agency of the United States Government.  Neither the United States    */
/*   Government nor the University of California nor any of their          */
/*   employees, makes any warranty, express or implied, or assumes any     */
/*   liability or responsibility for the accuracy, completeness, or        */
/*   usefulness of any information, apparatus, product, or process         */
/*   disclosed, or represents that its specific commercial products,       */
/*   process, or service by trade name, trademark, manufacturer, or        */
/*   otherwise, does not necessarily constitute or imply its               */
/*   endorsement, recommendation, or favoring by the United States         */
/*   Government or the University of California. The views and opinions    */
/*   of the authors expressed herein do not necessarily state or reflect   */
/*   those of the United States Government or the University of            */
/*   California, and shall not be used for advertising or product          */
/*   endorsement purposes.                                                 */
/*                                                                         */
/*   Permission to use, copy, modify and distribute this software and its  */
/*   documentation for any non-commercial purpose, without fee, is         */
/*   hereby granted, provided that the above copyright notice and this     */
/*   permission notice appear in all copies of the software and            */
/*   supporting documentation, and that all UC LLNL identification in      */
/*   the user interface remain unchanged.  The title to copyright LLNL     */
/*   XDIR shall at all times remain with The Regents of the University     */
/*   of California and users agree to preserve same. Users seeking the     */
/*   right to make derivative works with LLNL XDIR for commercial          */
/*   purposes may obtain a license from the Lawrence Livermore National    */
/*   Laboratory's Technology Transfer Office, P.O. Box 808, L-795,         */
/*   Livermore, CA 94550.                                                  */
/*                                                                         */
/***************************************************************************/
/***************************************************************************/

#include <Xm/Xm.h>
#include <string.h>
#include "xdir.h"
#include "list.h"

struct {
	int host;
	struct dirwin_st *src_dirwin;
	struct dirwin_st *snk_dirwin;
	char *snk_dir;
	struct entry_link *head;
	struct entry_link *head_src_dirs;
	int error;
} move_ctrl;

static char *not_implemented_msg = "Moving entry(s) from a host to another host not allowed.\nConsider copying entry(s) instead.";

static char *msg1 = "Unable to redisplay source directory.\n\nSource directory window might be out of date.";

static char *msg2 = "Unable to redisplay sink directory.\n\nSink directory window might be out of date.";

extern int inquire_on_move;
extern XtAppContext app;
extern int beep_when_ops_done;
extern int diagnostics;
extern struct st_host_info hinfo[];

int cb_move_entries();
char *entry_to_rel_path();
struct entry_info *rel_path_to_entry();
struct ref_st *create_reference_subtree();
char *entry_to_full_path();
char *merge_paths();
char *equivalent_file_path();
char **path_to_links();


/*
 * init_move - Initialize the move of selected entries in directory
 *             window "src_dirwin" to directory window "snk_dirwin".  If
 *             "einfo" is non-NULL then move the entries into the
 *             directory represented by "einfo".
 */
init_move(src_dirwin, snk_dirwin, einfo)
struct dirwin_st *src_dirwin;
struct dirwin_st *snk_dirwin;
struct entry_info *einfo;
{
	struct entry_link *head;
	struct entry_link *tail;
	struct entry_link *ptr;
	char *temp_snk_dir;
	char *snk_dir;
	int retval;
	int i;
	int j;
	int level;
	int need_to_update_entire_source;

	/* Sanity check */
	if (src_dirwin == snk_dirwin && einfo == NULL)
		fatal_error("Bug in init_move()");

	/* Is operation implemented */
	if (src_dirwin->host != snk_dirwin->host) {
		warn(not_implemented_msg, snk_dirwin->w_shell);
		return;
	}

	/* Ask the user if okay to move */
	if (inquire_on_move) {
		if (!verify_selection(src_dirwin, snk_dirwin,
				"Do you really want to move these items?"))
			return;
	}

    /* This might take some time */
    use_busy_cursor();

	/* Determine true sink directory path */
	if (einfo) {
		temp_snk_dir = entry_to_full_path(einfo);
	} else 
		temp_snk_dir = XtNewString(snk_dirwin->dirname);
    retval = cd(snk_dirwin->host, temp_snk_dir, True, &snk_dir);
	XtFree(temp_snk_dir);
    if (retval == -1) {
		restore_prev_cursor();
        record_and_alert("Unable to move entry(s).", snk_dirwin->w_shell);
        return;
    } else if (retval == -3) {
		restore_prev_cursor();
		lost_connection(snk_dirwin->host, snk_dirwin);
		return;
	}

    /* Pop up move monitor */
    show_mp_monitor(snk_dirwin, "Now Moving:");

    /* Disable controls */
    enable_controls(False);

    /* Form list of selected items to move */
    head = NULL;
    tail = NULL;
	for (i=0; i<src_dirwin->nentries; i++)
		if (src_dirwin->entries[i].level == 0)
		    postorder_selected_items(&src_dirwin->entries[i], &head, &tail);

	/* Set up move control block */
	move_ctrl.host = src_dirwin->host;
	move_ctrl.src_dirwin = src_dirwin;
	move_ctrl.snk_dirwin = snk_dirwin;
	move_ctrl.snk_dir = snk_dir;
	move_ctrl.head = head;
	move_ctrl.head_src_dirs = NULL;
	move_ctrl.error = False;

	/* Make list of src dirs to update after move. NULL path means update all */
	if (src_dirwin->layout == TREE) {
		need_to_update_entire_source = False;
		for (i=0; i<src_dirwin->nentries; i++)
			if (src_dirwin->entries[i].level == 0 
					&& src_dirwin->entries[i].state == SELECTED) {
				need_to_update_entire_source = True;
				break;
			}
		if (snk_dirwin == src_dirwin && einfo == NULL)
			need_to_update_entire_source = True;
		if (need_to_update_entire_source == False) {
			i = 0;
			while (i<src_dirwin->nentries) {
				if (src_dirwin->entries[i].type == DIRECTORY_TYPE) {
					level = src_dirwin->entries[i].level;
					if (einfo == &src_dirwin->entries[i] && einfo->expanded) {
						ptr = XtNew(struct entry_link);
						ptr->entry =
							entry_to_rel_path(einfo);
						ptr->next = move_ctrl.head_src_dirs;
						move_ctrl.head_src_dirs = ptr;
						for (i++; i<src_dirwin->nentries; i++)
							if (src_dirwin->entries[i].level <= level)
								break;
						continue;
					}
					for (j=i+1; j<src_dirwin->nentries; j++) {
						if (src_dirwin->entries[j].level <= level)
							break;
						if (src_dirwin->entries[j].level == level+1
								&& src_dirwin->entries[j].state == SELECTED) {
							ptr = ptr = XtNew(struct entry_link);
							ptr->entry =
								entry_to_rel_path(&src_dirwin->entries[i]);
							ptr->next = move_ctrl.head_src_dirs;
							move_ctrl.head_src_dirs = ptr;
							for (i=j+1; i<src_dirwin->nentries; i++)
								if (src_dirwin->entries[i].level <= level) {
									i--;
									break;
								}
							break;
						}
					}
				}
				i++;
			}
		}
	}

	/* Perform move */
	XtAppAddWorkProc(app, (XtWorkProc)cb_move_entries, NULL);
}


/*
 * cb_move_entries - Work proc to actually move directory entries.
 */
cb_move_entries()
{
    struct entry_link *ptr;
	char *src_path;
	char *snk_path;
	int retval;
	char msg[MAXPATHLEN+40];
	char *temp_path;
	char **links;
	char **link_ptr;
	int nlinks;

    /* Are we done? */
    if (move_ctrl.head == NULL) {
		if (move_ctrl.error)
			record_and_alert("At least one move failed.",
				move_ctrl.snk_dirwin->w_shell);
        goto done;
	}

	/* Did user push abort button? */
	if (mp_abort_requested())
		goto done;

	/* Let user know where we are */
	update_mp_monitor(move_ctrl.head->entry);

	/* Create source path */
	temp_path = merge_paths(hinfo[move_ctrl.src_dirwin->host].system,
		move_ctrl.src_dirwin->dirname, move_ctrl.head->entry);
	src_path = equivalent_file_path(hinfo[move_ctrl.src_dirwin->host].system,
		temp_path);
	XtFree(temp_path);

	/* Create sink path */
	links = path_to_links(hinfo[move_ctrl.snk_dirwin->host].system,
		move_ctrl.head->entry);
	nlinks = 0;
	link_ptr = links;
	while (*link_ptr++)
		nlinks++;
	temp_path = merge_paths(hinfo[move_ctrl.snk_dirwin->host].system,
		move_ctrl.snk_dir, links[nlinks-1]);
	release_path_links(links);
	snk_path = equivalent_file_path(hinfo[move_ctrl.snk_dirwin->host].system,
		temp_path);
	XtFree(temp_path);

	/* Move entry */
	if (move_ctrl.host == LOCAL)
		retval = local_rename(src_path, snk_path);
	else
		retval = remote_rename(move_ctrl.host, src_path, snk_path);
	XtFree(src_path);
	XtFree(snk_path);
	switch (retval) {
	case -3:
		goto lost;
	case -1:
		sprintf(msg, "Unable to move %s", move_ctrl.head->entry);
		record_warning(msg);
		move_ctrl.error = True;
		break;
	case 0:
		if (diagnostics >= NORMAL) {
			sprintf(msg, "*** Successfully moved:  %s\n",move_ctrl.head->entry);
			write_log(msg);
		}
	}
	ptr = move_ctrl.head;
	move_ctrl.head = move_ctrl.head->next;
	XtFree(ptr->entry);
	XtFree((char *)ptr);
	return False;

done:

	if (mp_abort_requested()) {
		hide_abort_dialog();
		write_log("*** Move(s) aborted.");
	} else
		hide_mp_monitor();
	enable_controls(True);
	update_move_displays();
	clean_up_move_ctrl();
	update_dir_controls();
    if (beep_when_ops_done)
        beep();
    restore_prev_cursor();
    return True;

lost:

    clean_up_move_ctrl();
	if (mp_abort_requested()) {
		hide_abort_dialog();
		write_log("*** Move(s) aborted.");
	} else
		hide_mp_monitor();
    enable_controls(True);
    clear_selected_entries();
    restore_prev_cursor();
    lost_connection(move_ctrl.host, move_ctrl.snk_dirwin);
    use_busy_cursor();
    update_dir_controls();
    if (beep_when_ops_done)
        beep();
    restore_prev_cursor();
    return True;
}


/*
 * update_move_displays - Call after move operation is complete to update
 *                        source and sink directory displays.
 */
update_move_displays()
{
	struct entry_info *einfo;
	struct ref_st *ref_tree;
	struct entry_link *ptr;
	struct dirwin_st *snk_dirwin;
	int indx;
	char *anchor_rel_path;
	int retval;
	
	/* Update source display */
	if ((move_ctrl.src_dirwin->layout == TREE) && move_ctrl.head_src_dirs) {
		indx = move_ctrl.src_dirwin->first_visible_row;
		anchor_rel_path =
			entry_to_rel_path(&move_ctrl.src_dirwin->entries[indx]);
		ptr = move_ctrl.head_src_dirs;
		while (ptr) {
			einfo = rel_path_to_entry(move_ctrl.src_dirwin, ptr->entry);
			if (einfo) {
				ref_tree = create_reference_subtree(einfo);
				collapse_subtree(&einfo, False);
				retval = re_expand_subtree(&einfo, ref_tree);
				release_reference_subtree(ref_tree);
				switch (retval) {
				case -1:
					record_and_alert(msg1, move_ctrl.snk_dirwin);
					break;
				case -3:
					XtFree(anchor_rel_path);
					lost_connection(move_ctrl.src_dirwin->host,
						move_ctrl.snk_dirwin);
					goto sink;
				}
			}
			ptr = ptr->next;
		}
		calc_dirwin_geometry(move_ctrl.src_dirwin);
		set_first_visible_entry(move_ctrl.src_dirwin, anchor_rel_path);
		XtFree(anchor_rel_path);
	} else {
		retval = display_dir(move_ctrl.src_dirwin->host, move_ctrl.src_dirwin,
			move_ctrl.src_dirwin->dirname, True, False);
		switch (retval) {
		case -1:
			record_and_alert(msg1, move_ctrl.snk_dirwin);
			break;
		case -3:
			lost_connection(move_ctrl.src_dirwin->host, move_ctrl.snk_dirwin);
		}
	}

sink:

	/* Update sink display */
	if (is_dir_displayed(move_ctrl.host, move_ctrl.snk_dir, &snk_dirwin,
			&einfo) && (snk_dirwin != move_ctrl.src_dirwin)) {
		if (einfo) {
			indx = snk_dirwin->first_visible_row;
			anchor_rel_path = entry_to_rel_path(&snk_dirwin->entries[indx]);
			ref_tree = create_reference_subtree(einfo);
			collapse_subtree(&einfo, False);
			retval = re_expand_subtree(&einfo, ref_tree);
			release_reference_subtree(ref_tree);
			switch (retval) {
			case -1:
				record_and_alert(msg2, move_ctrl.snk_dirwin);
				break;
			case -3:
				lost_connection(move_ctrl.snk_dirwin->host,
					move_ctrl.snk_dirwin);
				XtFree(anchor_rel_path);
				return;
			}
			calc_dirwin_geometry(snk_dirwin);
			set_first_visible_entry(snk_dirwin, anchor_rel_path);
			XtFree(anchor_rel_path);
		} else {
			retval = display_dir(snk_dirwin->host, move_ctrl.snk_dirwin,
				move_ctrl.snk_dir, True, False);
			switch (retval) {
			case -1:
				record_and_alert(msg2, move_ctrl.snk_dirwin);
				break;
			case -3:
				lost_connection(move_ctrl.snk_dirwin->host,
					move_ctrl.snk_dirwin);
			}
		}
	}
}


/*
 * clean_up_move_ctrl - Free up memory used by the "Move" control block
 *                      data structure.
 */
clean_up_move_ctrl()
{
	release_linked_list(&move_ctrl.head);
	release_linked_list(&move_ctrl.head_src_dirs);
	XtFree(move_ctrl.snk_dir);
}

